use eyre::Result;
use log::error;
use std::env;
use std::io::Write;
use std::process::exit;

use init::cmd;

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<()> {
    let mut log_builder = env_logger::builder();

    let call_name = env::args().next().unwrap_or("init".into());
    let call_name = (call_name.rsplit_once('/').map(|(_, n)| n)).unwrap_or(call_name.as_str());

    if call_name == "init" {
        let mut lw = LogWriter::new();
        lw.push(std::io::stderr());
        lw.push(std::fs::File::create("/var/log/init.log")?);

        log_builder.target(env_logger::Target::Pipe(Box::new(lw)));
    }

    log_builder
        .format_timestamp_millis()
        .format_target(false)
        .filter_level(log::LevelFilter::Info)
        .write_style(env_logger::WriteStyle::Always)
        .init();

    match call_name {
        "init" => cmd::init::run().await,
        "init-version" => cmd::version::run(),
        "init-input" => cmd::init_input::run().await,

        _ => {
            error!("invalid call name: {call_name:?}");
            exit(1);
        }
    };

    Ok(())
}

struct LogWriter {
    targets: Vec<Box<dyn Write + Send + 'static>>,
}

impl LogWriter {
    pub fn new() -> Self {
        Self {
            targets: Vec::new(),
        }
    }

    pub fn push<W: Write + Send + 'static>(&mut self, w: W) {
        self.targets.push(Box::new(w));
    }
}

impl Write for LogWriter {
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        for target in self.targets.iter_mut() {
            target.write_all(buf)?;
        }
        Ok(buf.len())
    }

    fn flush(&mut self) -> std::io::Result<()> {
        for target in self.targets.iter_mut() {
            target.flush()?;
        }
        Ok(())
    }
}
